// Montecarlo.cpp : Defines the entry point for the console application.
//


#include <iostream>
using namespace std;
//#include <tchar.h>
//#include <windows.h>
#include <math.h>
#include <GL/gl.h>
#include "GL/glut.h"
#include "ringlet/pkgs/R3Graphics/R3Graphics.h"
#include "Vector.h"
using namespace CDLib;
#include <vector>
using std::vector;
#include <string>
using std::string;
#include "Noise.h"



// RING include files



// Program arguments
double globalstate = 0; double globali=0, globalj=0;
static char *scenename = NULL;
static char *imagename = NULL;
static const int imagewidth = 512; //320
static const int imageheight = imagewidth/4*3;
static const int actualwidth = 512;
static const int actualheight = 512;
static RNBoolean print_verbose = FALSE;
int grid_dimension = 3;
int render_passes = 1;

int linesrendered=0;
int linesdrawn=0;

//HANDLE pass_barrier_event;

bool display = true;
bool display_tex = true;

#define CDVECTOR(x) (*(Vector3f*)&x)
#define R3VECTOR(x) (*(R3Vector*)&x)

GLfloat pixelsums[imagewidth][imageheight][4] = {0};
GLfloat pixelvalues[actualwidth][actualheight][4] = {0};
int pixelcounts[imagewidth][imageheight] = {0};

int win;
//CRITICAL_SECTION lock;
bool changed = true;
R2Image *image = NULL;
R3Scene *scene = NULL;
unsigned int texId=0;

bool _isnan(float x)
{
	return x != x;
}

float threshold(float x, float a=0.0f, float b=1.0f)
{
	if( x < a ) x = a;
	else if( x > b ) x = b;
	return x;
}

string NextFile(string base)
{
	char name[255];
	int count=0;
	FILE *file;

	while(true)
	{
		sprintf(name, "%s%03d.bmp", base.c_str(), count);
		if( (file=fopen(name, "r")) == NULL )
			break;
		
		fclose(file);
		count++;
	}
	
	return name;	
}

void WriteImage(const char* name = NULL)
{
	if( name == NULL )
		name = imagename;

	R2Image* blah = new R2Image(imagewidth, imageheight);
	for(int i=0; i<imagewidth; i++)
		for(int j=0; j<imageheight; j++)
		{
			RNRgb rgb;
			for(int k=0; k<3; k++)
				rgb[k] = threshold(pixelvalues[i][j][k], 0.0f, 1.0f);;
			
			blah->SetPixelRGB(i,j,rgb);
		}

	blah->Write(name);
	
	delete blah;
}

inline double safesqrt(double x)
{
	return x < 0.0 ? 0.0 : sqrt(x);
}

void sample_hemisphere(R3Vector& n, R3Vector& out)
{
	double phi = frand(0, 2*M_PI);
	double z = frand();
	
	double sinTheta, sinPhi, cosPhi;
	if( true )
	{
		double theta = acos(z);
		sinTheta = sin(theta);
		sinPhi = sin(phi);
		cosPhi = cos(phi);
	}

	out[0] = sinTheta*cosPhi;
	out[1] = sinTheta*sinPhi;
	out[2] = z;

	if( out[0]*n[0] + out[1]*n[1] + out[2]*n[2] < 0 )
	{
		out[0] = -out[0];
		out[1] = -out[1];
		out[2] = -out[2];
	}
}

vector<R3Node*> lights;

void Illuminate(const R3Point& point, const R3Vector& normal, 
				const R3Vector& incoming, const R3Vector& outgoing, 
				const R3Node* node, R3Vector& out )
{
	R3Brdf &brdf = *(R3Brdf*)(node->Material()->Brdf());

	//Compute the diffuse factor first
	for(int i=0; i<3; i++)
		out[i] = brdf.Diffuse()[i] / M_PI;

	//If it is specular, add in the specular component
	if( node->Material()->IsSpecular() )
	{
		double n = brdf.Shininess();
		R3Vector reflect = -2 * (incoming.Dot(normal)) * normal + incoming;
		double cosalpha = std::max((float)reflect.Dot(outgoing), 0.0f);		
		double specfactor = ((n+2)/(2*M_PI)) * pow(cosalpha, n);
			
		for(int i=0; i<3; i++)
		{
			double specular = brdf.Specular()[i] * specfactor;
            out[i] += specular;
		}
	}

	return;
}

inline float Luminance(const RNRgb& c)
{
     // Return luminance
     return 0.30 * c[0] + 0.59 * c[1] + 0.11 * c[2];
}

bool Roulette(const R3Vector& w)
{
	if( w[0]*w[0] + w[1]*w[1] + w[2]*w[2] < 0.001 )
		return true;
	else
		return false;
}

struct Triangle
{
	R3Vector vtx[3];
	R3Vector plane_normals[3];
	double angles[3];
	double lengths[3];
	double area;
};

int prev(int i)
{
	i -= 1;
	if( i == -1 )
		i = 2;
	return i;
}

void ComputeTriangleInfo(Triangle& out)
{
	//Compute each of the plane normals
	for(int i=0; i<3; i++)
	{
		//out.plane_normals[i] = (out.vtx[i] - center.Vector());
		//out.plane_normals[i].Cross(out.vtx[(i+1)%3] - center.Vector());
		out.plane_normals[i] = (out.vtx[i]);
		out.plane_normals[i].Cross(out.vtx[(i+1)%3]);
		out.plane_normals[i].Normalize();
	}

	//Compute each of the angles between the planes
	for(int i=0; i<3; i++)
		out.angles[i] = M_PI - acos( threshold(out.plane_normals[i].Dot(out.plane_normals[prev(i)]),-1.0f,1.0f) );
	assert( !_isnan(out.angles[0]) && !_isnan(out.angles[1]) && !_isnan(out.angles[2]) );

	//Compute the area
	out.area = out.angles[0] + out.angles[1] + out.angles[2] - M_PI;
	
	//Compute the arc lengths
	//R3Vector diff = out.vtx[1] - out.vtx[2];
	//double len = sqrt(diff.Dot(diff));

	out.lengths[0] = acos( threshold(out.vtx[1].Dot(out.vtx[2]),-1.0f,1.0f) );
	out.lengths[1] = acos( threshold(out.vtx[0].Dot(out.vtx[2]),-1.0f,1.0f) );
	out.lengths[2] = acos( threshold(out.vtx[0].Dot(out.vtx[1]),-1.0f,1.0f) );
}

void ProjectToSphere(const R3Point center, const R3Triangle* in, Triangle& out)
{
	//R3Point pts[3] = { R3Point(-1,0.5,0.5), R3Point(0.5,1,0.5), R3Point(0.5,0.5,1) };
	//R3Point pts[3] = { R3Point(0.5,-1,0.5), R3Point(0.5,1,0.5), R3Point(0.5,0.5,1) };
	//R3Point pts[3] = { R3Point(-1,0.5,0.5), R3Point(0.5,-1,0.5), R3Point(0.5,-1,0.7) };

	//Project the vectices onto the sphere
	for(int i=0; i<3; i++)
	{
		//const R3Point &v = pts[i];
		const R3Point &v = in->Vertex(i)->Position();
		out.vtx[i] = v-center;
		out.vtx[i].Normalize();
	} 

	ComputeTriangleInfo(out);
}

int VerticesInHemisphere(Triangle& T, R3Vector& n)
{
	int count=0;
	for(int i=0; i<3; i++)
		if( T.vtx[i].Dot(n) > 0.0f ) 
			count++;
	return count;
}

/*
void SampleTriangle(const R3Triangle* T, R3Vector& out, double e1 = -1, double e2 = -1)
{
	double s = (e1 == -1 ? sqrt(frand()) : sqrt(e1));
	double t = (e2 == -1 ? e2 : frand());

	out = R3VECTOR( 
		(1-s)*T->Vertex(0)->Position() +
		s*(1-t)*T->Vertex(1)->Position() +
		s*t*T->Vertex(2)->Position()    );
	return;
}*/

bool Visible(R3Vector x, R3Vector y)
{
	R3Vector d = y - x;
	R3Ray ray( *(R3Point*)&x, d );

	R3Node* node;
	R3Vector point;
	R3Vector normal;
	if( !scene->Intersects(ray, &node, (R3Point*)&point) )
		return false;

	R3Vector d2 = point - y;
	if( abs(sqrt(d2.Dot(d2)) - 1.0) < 1e-4 )
		return true;
	else
		return false;
}

void SampleSphericalTriangle(Triangle& T, R3Vector& out)
{
	//Taken from Arvo95
	double eps= 0.0;
	double alpha = T.angles[0];
	R3Vector &A = T.vtx[0];
	R3Vector &B = T.vtx[1];
	R3Vector &C = T.vtx[2];
	double c = T.lengths[2];

	//Use one random variable to select the new area
	double Ah = frand(eps,1-eps) * T.area;

	//Save the sine and cosine of angle phi
	double s = sin(Ah - alpha);
	double t = cos(Ah - alpha);

	//Compute the pair u,v that determines BetaHat
	double u = t - cos(alpha);
	double v = s + sin(alpha)*cos(c);

	//Let q be the cos of the new edge length bh
	double q = ((v*t - u*s)*cos(alpha) - v) /
		       ((v*s + u*t)*sin(alpha));

	//COmpute the third vertex of the subtriangle
	R3Vector CA = C - C.Dot(A)*A;
	CA.Normalize();
	//assert(1-q*q >= 0);
	R3Vector Ch = q*A + safesqrt(1-q*q)*CA;

	//Use the other random variable to select cos theta
	double z = 1 - frand(eps,1-eps)*(1 - Ch.Dot(B));

	//Construct the corresponding point on the sphere
	R3Vector CB = Ch - Ch.Dot(B)*B;
	CB.Normalize();
	out = z*B + safesqrt(1-z*z)*CB;
	out.Normalize();
	assert( abs(out.Length() - 1.0) < 1e-6 );
}

void SamplePhongLobe(R3Vector& reflect, double n, R3Vector& out)
{
	R3Vector &W = reflect;

	double e1 = frand();
	double e2 = frand();

	out[0] = sqrt(1 - pow(e1, 2/(n+1))) * cos(2*M_PI*e2);
	out[1] = sqrt(1 - pow(e1, 2/(n+1))) * sin(2*M_PI*e2);
	out[2] = pow(e1, 1/(n+1));

	R3Vector T = W;
	RNDimension d = T.MinDimension();
	T[d] = 1.0;

	R3Vector U = T;
	U.Cross(W);
	U.Normalize();

	R3Vector V = W;
	V.Cross(U);
	V.Normalize();

	R4Matrix R(
		U[0], V[0], W[0], 0,
		U[1], V[1], W[1], 0,
		U[2], V[2], W[2], 0,
		0,    0,    0,    1);
	R3Affine R2(R);

	R2.Apply(out);
}

void ClipSphericalTriangle(R3Vector& n, Triangle& T, vector<Triangle>& out)
{
	if( T.area < 1e-6 )
		return;

	int vtxin = VerticesInHemisphere(T, n);
	switch(vtxin)
	{
	case 0:
		return;
	case 1:
		{
			//Rearrainge so that the inside vertex is vtx[0];
			R3Vector *vtx[3]; 
			if( T.vtx[0].Dot(n) > 0 )
			{
				vtx[0] = &T.vtx[0];
				vtx[1] = &T.vtx[1];
				vtx[2] = &T.vtx[2];
			}
			else if( T.vtx[1].Dot(n) > 0 )
			{
				vtx[0] = &T.vtx[1];
				vtx[1] = &T.vtx[0];
				vtx[2] = &T.vtx[2];
			}
			else
			{
				vtx[0] = &T.vtx[2];
				vtx[1] = &T.vtx[0];
				vtx[2] = &T.vtx[1];
			}
			
			Triangle U;
			U.vtx[0] = *vtx[0];

			//Find the two new vertices
			R3Vector& A = *vtx[0];
			for(int i=1; i<=2; i++)
			{
				R3Vector& B = *vtx[i];
				U.vtx[i] = A - (n.Dot(A) / (n.Dot(B-A))) * (B-A);
				U.vtx[i].Normalize();
			}

			ComputeTriangleInfo(U);
			out.push_back(U);
			return;
		}
	case 2:
		{
			//Rearrainge so that the outside vertex is vtx[0];
			R3Vector *vtx[3]; 
			if( T.vtx[0].Dot(n) < 0 )
			{
				vtx[0] = &T.vtx[0];
				vtx[1] = &T.vtx[1];
				vtx[2] = &T.vtx[2];
			}
			else if( T.vtx[1].Dot(n) < 0 )
			{
				vtx[0] = &T.vtx[1];
				vtx[1] = &T.vtx[0];
				vtx[2] = &T.vtx[2];
			}
			else
			{
				vtx[0] = &T.vtx[2];
				vtx[1] = &T.vtx[0];
				vtx[2] = &T.vtx[1];
			}
			
			R3Vector newvtx[2];

			//Find the two new vertices
			R3Vector& A = *vtx[0];
			for(int i=1; i<=2; i++)
			{
				R3Vector& B = *vtx[i];
				newvtx[i-1] = A - (n.Dot(A) / (n.Dot(B-A))) * (B-A);
				newvtx[i-1].Normalize();
			}

			//Set the vertices for the two new triangles
			Triangle U, V;
			U.vtx[0] = *vtx[1];
			U.vtx[1] = newvtx[0];
			U.vtx[2] = *vtx[2];
			
			V.vtx[0] = newvtx[0];
			V.vtx[1] = newvtx[1];
			V.vtx[2] = *vtx[2];

			ComputeTriangleInfo(U);
			ComputeTriangleInfo(V);

			out.push_back(U);
			out.push_back(V);
			return;
		}
	case 3:
		out.push_back(T);
		return;
	}

	assert(false);
	return;
}

bool HitsLight(R3Point& p, R3Vector& d, int light)
{
	R3Ray ray( p + 0.01*d , d);
	R3Node* node;
	R3Point point;
	R3Vector normal;
	RNScalar t;
	if( !scene->Intersects(ray, &node, &point, &normal, &t) || t < 1e-7 )
		return false;
	
	if( node->Material()->IsEmissive() )
	{
		if( light == -1 )
			return true;
		else
		{
			int idx;
			for(int i=0; i<lights.size(); i++)
				if( lights[i] == node )
					idx = i;
			
			if( idx == light )
				return true;
			else return false;
		}
	}
	else 
		return false;
}

//A: 1.0472 angle1: 
//We need several things:
//1. For triangle T, project to heimsphere (ProjectToSphere)
//2. For spherical triangle T, determine if T is entirely on the hemisphere (VerticesInHemisphere)
//3. For spherical triangle T, determine the area (ProjectToSphere)
//4. For spherical triangle T, select a random point on T (SampleSphericalTriangle)
//5. For point P, determine if the path to P is unoccluded 
//6. Randomly select triangles based on projected solid angle
//7. Select a uniform random direction in hemisphere (sample_hemisphere)
//8. Determine if the point hits a triangle within hemisphere

//What are the different cases?

//Emission
//If the shape is totally emissive, return the emission
//If the shape is partially emissive, set P_emit = 0.9, randomly emit
//If the shape is non-emissive, set P_emit = 0.0

//Direct Illumination
//If no light exists, do not shoot to light, P_direct = 0.0, P_indirect = 1.0 * (1-P_emit)
//If light exists, P_direct = 2pi/sa(light) * (1-P_emit)
//   P_indirect = 2pi/(2pi-sa(light)) * (1-P_emit), rgb += TracePath(), samples++

//Indirect Illumination
//If the shape is diffuse, P_diffuse = P_indirect * 1/pi * cosND, rgb += TracePath(), samples++


vector<Triangle> triangles;
void TracePath(const R3Point& p, const R3Vector& d, const R3Vector& weightin, int bounces, R3Vector& rgbout)
{
	//Construct ray and fire it at the scene
	//Hack it slightly to avoid intersection with the last object
	R3Ray ray( p + 0.01*d , d);
	R3Point point;
	R3Vector normal;
	R3Node* node;
	RNScalar t;

	globalstate = 1.5;

	rgbout[0] = rgbout[1] = rgbout[2] = 0.0;
	if( bounces == 15 || !scene->Intersects(ray, &node, &point, &normal, &t) || t < 1e-7 )
		return;
	assert( !_isnan(point.X()) && !_isnan(point.Y()) && !_isnan(point.Z()) );


	globalstate = 2;
	//If the BRDF for the hit surface is emissive, randomly emit
	double P_emit = 0.0;
	R3Brdf &brdf = *(R3Brdf*)(node->Material()->Brdf());
	if( node->Material()->IsEmissive() )
	{
		if( node->Material()->IsDiffuse() || node->Material()->IsSpecular() )
            P_emit = 0.9;
		else 
			P_emit = 1.0;
	}

	if(	P_emit == 1.0 || (P_emit > 0.0 && frand() < P_emit) )
	{
		for(int k=0; k<3; k++) rgbout[k] = weightin[k] * brdf.Emission()[k] / P_emit;
		if( bounces == 0 && rgbout[0] < 0.7 )
			printf("Error in pixel\n");
		return;
	}

	
	//Play russian roulette to see if we stop now
	bool roulette=false;
	if( Roulette(weightin) )
	{
		rgbout[0] = rgbout[1] = rgbout[2] = 0.0;
		//roulette = true;
		return;
	}

//	double P_roulette = 0.5;

	globalstate = 3;

	//Otherwise, we compute the reflected light
	R3Vector pathrgb;
	R3Vector dir; 
	
	//Build a vector with all of the clipped visible lights
	vector<Triangle> tris;
	for(int i=0; i<lights.size(); i++)
	{
		R3Shape *S = (R3Shape*)lights[i]->Shape();
		R3Triangle *T = (R3Triangle*)S;
		Triangle ST;
		ProjectToSphere(point, T, ST);
		ClipSphericalTriangle(normal, ST, tris);
	}

	//Find the total area
	double sa = 0.0;
	for(int i=0; i<tris.size(); i++)
		sa += tris[i].area;

	globalstate = 4;

	/*
	//Decide if we will shoot to a light or a non-light by a random chance
	double P_shoot = 0.5;
	if( tris.size() == 0 )
		P_shoot = 0.0;
	else if( roulette == true )
		P_shoot = 1.0;
	*/

	//Handle the special case of roulette but no visible light
	/* if( roulette ) //&& tris.size() == 0 )
	{
		rgbout[0] = rgbout[1] = rgbout[2] = 0.0;
		return;
	} */

	//Decide if we are going to sample the specular lobe
	int samples=0;
	/* if( node->Material()->IsSpecular()  )
	{
			samples++;
			double k_d = brdf.Diffuse().Length();
			double k_s = brdf.Specular().Length();
						
	} */

	R3Vector weightout, reflectance;
	//Sample the diffuse lobe
	
	//Now look for a light to shoot to, weighted by area
	int thelight = -1;
	double P_direct = 0.0, P_indirect = (1-P_emit), P_diffuse, P_specular;
	
	if( tris.size() > 0 )
	{
		samples++;
		
		//do
		//{
		double alpha = frand(0,sa);
		double beta=0.0;
		thelight = 0;
		for(int i=0; i<tris.size(); i++)
		{
			if( alpha < beta + tris[i].area )
			{
				thelight = i;
				break;
			}
			else
				beta += tris[i].area;
		}
		assert( thelight != -1 );

		//Sample a point on that spherical triangle
		SampleSphericalTriangle(tris[thelight], dir);
		//}
		//while( !HitsLight(point, dir, thelight) );


		//Compute the probability density
		//P_ray = (1-P_emit) * P_shoot * (2*M_PI/sa);
		P_direct = (1-P_emit) * (2*M_PI/sa); //this may need to consider the weighted average
		P_indirect = (1-P_emit) * ((2*M_PI) / (2*M_PI-sa));

		//if( HitsLight(point, dir, thelight) )
		//{
			Illuminate(point, normal, d, dir, node, reflectance);
			for(int k=0; k<3; k++)	weightout[k] = weightin[k] * reflectance[k] * normal.Dot(dir) / P_direct;

			TracePath(point, dir, weightout, bounces+1, pathrgb);
			for(int k=0; k<3; k++) rgbout[k] += pathrgb[k];
		//}
	}
	

	globalstate = 8;

	if( node->Material()->IsSpecular() )
	{
		R3Vector reflect;
		double shiny = brdf.Shininess();
		do
		{
			reflect = -2 * (d.Dot(normal)) * normal + d;
			SamplePhongLobe(reflect, shiny, dir);
		} while( HitsLight(point, dir, -1) );

		double cosalpha = std::max((float)reflect.Dot(dir), 0.0f);		
		double P_specular = P_indirect * ((shiny + 1)/(2*M_PI))*pow(cosalpha, shiny);
		
		if( P_specular > 1e-6 )
		{
			Illuminate(point, normal, d, dir, node, reflectance);
			for(int k=0; k<3; k++)	weightout[k] = weightin[k] * reflectance[k] /* * normal.Dot(dir)*/ / P_specular;

			TracePath(point, dir, weightout, bounces+1, pathrgb);
			rgbout += pathrgb;
		}
		//for(int k=0; k<3; k++) rgbout[k] += pathrgb[k];
	}

	//Now shoot at a nonlight source
	do 	
	{
		//Sample randomly in the hemisphere
		sample_hemisphere(normal, dir);

		//If it hits a light, repeat the process
	} while( HitsLight(point, dir, -1) );

	//Compute the probability density
	//P_ray = (1-P_emit) * (1-P_shoot) * ((2*M_PI) / (2*M_PI-sa));
	
	globalstate = 9;

	//P_diffuse = P_indirect*M_PI;
	P_diffuse = P_indirect;

	Illuminate(point, normal, d, dir, node, reflectance);
	for(int k=0; k<3; k++)	weightout[k] = weightin[k] * reflectance[k] * normal.Dot(dir) / (P_diffuse);

	globalstate = 10;

	TracePath(point, dir, weightout, bounces+1, pathrgb);
	
	globalstate = 11;

	for(int k=0; k<3; k++) rgbout[k] += pathrgb[k];
	samples++;

	globalstate = 12;

	//for(int k=0; k<3; k++) rgbout[k] /= samples;
	
	globalstate = 13;
	

	//Send it off to the new point, with new weights


	//Illuminate(point, normal, d, dir, node, reflectance);
	//for(int k=0; k<3; k++)	weightout[k] = weightin[k] * reflectance[k] * normal.Dot(dir) / (P_ray);

	//TracePath(point, dir, weightout, bounces+1, pathrgb);
	//for(int k=0; k<3; k++) rgbout[k] += pathrgb[k];

	return;
}

void TracePathSimple(const R3Point& p, const R3Vector& d, const R3Vector& weightin, int bounces, R3Vector& rgbout)
{
	//Construct ray and fire it at the scene
	R3Ray ray( p + 0.01*d , d);
	R3Point point;
	R3Vector normal;
	R3Node* node;
	RNScalar t;

	globalstate = 1.5;

	rgbout[0] = rgbout[1] = rgbout[2] = 0.0;
	if( bounces == 5 || !scene->Intersects(ray, &node, &point, &normal, &t) || t < 1e-7 )
		return;
	assert( !_isnan(point.X()) && !_isnan(point.Y()) && !_isnan(point.Z()) );

	//Compute the emissive component
	double P_emit = 0.0;
	R3Brdf &brdf = *(R3Brdf*)(node->Material()->Brdf());
	if( node->Material()->IsEmissive() )
	{
		if( node->Material()->IsDiffuse() || node->Material()->IsSpecular() )
            P_emit = 0.9;
		else 
			P_emit = 1.0;
	}

	if(	P_emit == 1.0 || (P_emit > 0.0 && frand() < P_emit) )
	{
		for(int k=0; k<3; k++) rgbout[k] = weightin[k] * brdf.Emission()[k] / P_emit;
		if( bounces == 0 && rgbout[0] < 0.7 )
			printf("Error in pixel\n");
		return;
	}

	//Sample random ray to compute the diffuse component
	R3Vector dir, reflectance, weightout, pathrgb;
	sample_hemisphere(normal, dir);

	Illuminate(point, normal, d, dir, node, reflectance);
	for(int k=0; k<3; k++)	weightout[k] = weightin[k] * reflectance[k] * normal.Dot(dir) / (1-P_emit);

	TracePathSimple(point, dir, weightout, bounces+1, pathrgb);
	for(int k=0; k<3; k++) rgbout[k] += pathrgb[k];

}



	/*
	//Compute the solid angle for all lights
	if( frand() < 0.5 )
	{
		//Randomly select a light and a point on that light
		//do 		{
			int L = rand() % lights.size();
			R3Shape *S = (R3Shape*)lights[L]->Shape();
			R3Triangle *T = (R3Triangle*)S;

			Triangle ST;
			ProjectToSphere(point, T, ST);
			SampleSphericalTriangle(ST, dir);
		//} while( dir.Dot(normal) < 0.0 );
						
		if( ST.area < 1e-6 )
		{
			rgbout[0] = rgbout[1] = rgbout[2] = 0.0;
			return;
		}
		
		P_ray = (1- P_emit) * 0.5 * (2*M_PI/ ST.area);
		assert(P_ray > 0.0);
		
	}
	else
	{
		//Compute the random direction
		sample_hemisphere(normal, dir);
		//P_ray = (1- P_emit) * 0.5 * (2*M_PI/(2*M_PI-sa));
		P_ray = (1- P_emit) * 0.5;
		assert(P_ray > 0.0);
	}
	*/


void TracePathLight(const R3Point& p, const R3Vector& d, const R3Vector& weightin, int bounces, R3Vector& rgbout)
{
	//Construct ray and fire it at the scene
	R3Ray ray( p + 0.01*d , d);
	R3Point point;
	R3Vector normal;
	R3Node* node;
	RNScalar t;

	globalstate = 1.5;

	rgbout[0] = rgbout[1] = rgbout[2] = 0.0;
	if( bounces == 5 || !scene->Intersects(ray, &node, &point, &normal, &t) || t < 1e-7 )
		return;
	assert( !_isnan(point.X()) && !_isnan(point.Y()) && !_isnan(point.Z()) );

	//Compute the emissive component
	double P_emit = 0.0;
	R3Brdf &brdf = *(R3Brdf*)(node->Material()->Brdf());
	if( node->Material()->IsEmissive() )
	{
		if( node->Material()->IsDiffuse() || node->Material()->IsSpecular() )
            P_emit = 0.9;
		else 
			P_emit = 1.0;
	}

	if(	P_emit == 1.0 || (P_emit > 0.0 && frand() < P_emit) )
	{
		for(int k=0; k<3; k++) rgbout[k] = weightin[k] * brdf.Emission()[k] / P_emit;
		if( bounces == 0 && rgbout[0] < 0.7 )
			printf("Error in pixel\n");
		return;
	}

	//Decide to shoot to a light or a non-light
	if( frand() < 0.5 )
	{


	}

	//Sample random ray to compute the diffuse component
	R3Vector dir, reflectance, weightout, pathrgb;
	sample_hemisphere(normal, dir);

	Illuminate(point, normal, d, dir, node, reflectance);
	for(int k=0; k<3; k++)	weightout[k] = weightin[k] * reflectance[k] * normal.Dot(dir) / (1-P_emit);

	TracePathSimple(point, dir, weightout, bounces+1, pathrgb);
	for(int k=0; k<3; k++) rgbout[k] += pathrgb[k];

}

void RayThroughPixel(
	int i, int maxi, int j, int maxj, int s, int dim, 
	const R3Camera* camera, R3Vector &dir)
{
	//int dim = sqrt((double)maxs);
	float gap = 1.0f/(dim+1);
	float scale = 1.0f;

	//Get the indices of the sample in the subpixel grid
	//Remember, the grid does not go to the edges
	int x1 = s % dim;
	int y1 = s / dim;

	//Get the floating point position in the subpixel grid
	//Recenter so that they are relative to center
	float x2 = (((float)x1) * gap + gap - 0.5)*scale;
	float y2 = (((float)x1) * gap + gap - 0.5)*scale;
	
	//Jitter within radius of dim
	float x3 = frand(-gap*scale, gap*scale) + x2;
	float y3 = frand(-gap*scale, gap*scale) + y2;

	RNScalar dx = (2.0 * (i+x3) - maxi) / maxi;
	RNScalar dy = (2.0 * (j+y3) - maxj) / maxj;
	R3Point far_org = camera->Origin() + camera->Towards() * camera->Far();
	R3Vector far_right = camera->Right() * camera->Far() * tan(camera->XFOV());
	R3Vector far_up = camera->Up() * camera->Far() * tan(camera->YFOV());
	R3Point far_point = far_org + (far_right * dx) + (far_up * dy);

	//dir = *(R3Vector*)&far_point;
	dir = far_point.Vector() - camera->Origin().Vector();
	dir.Normalize();
}

static void Render(R3Scene *scene, R2Image *image)
{
	// Start statistics
	RNTime start_time;
	start_time.Read();

	// Get camera
	const R3Camera *camera = scene->Camera();
	if (!camera) RNAbort("There is no camera in the scene");

	bool pathtrace = true;

	//Find all the lights in the scene
	for(int i=0; i<scene->NNodes(); i++)
	{
		R3Node* node = (R3Node*)scene->Node(i);
		if( node->Material()->IsEmissive() )
			lights.push_back(node);
	}

	//Luminance 0.3 R 0.59 G 0.11 B
	//_controlfp(_EM_INEXACT | _EM_UNDERFLOW, _MCW_EM);

	// Generate ray through each pixel
	for(int p = 0; p < render_passes; p++)	{
		printf("Rendering pass %d... ", p);
		fflush(stdout);
		for (int i = 0; i < imagewidth; i++) {
			linesrendered = i;
			for (int j = 0; j < imageheight; j++) {

				double ij = i*image->Height() + j;

				//show_status( ij/(image->Width()*image->Height()) );

				// Find intersection
				R3Node *node;
				R3Point point;
				R3Vector normal;
				RNScalar t;

				RNRgb rgb;

				int dim=grid_dimension;
				int samples=dim*dim;
				R3Vector pixelrgb(0,0,0);

				for(int s=0; s<samples; s++)
				{
					globalstate = 0;
					
					R3Vector far_dir;
					RayThroughPixel(i, imagewidth, j, imageheight, s, dim, camera, far_dir);

					R3Vector pathrgb(0,0,0);
					R3Vector weight(1,1,1);

					TracePath(camera->Origin(), far_dir, weight, 0, pathrgb);
					globalstate = 14;
					for(int k=0; k<3; k++)
					{
						//assert( pathrgb[k] < 200 );
						//assert( pathrgb[k] > -200.0 );
						assert( !_isnan(pathrgb[k]) );
						pixelrgb[k] += threshold(pathrgb[k], 0.0f, 2.0f);
						//pixelrgb[k] += pathrgb[k];
					}
					globalstate = 15;
						//pixelrgb[k] += pathrgb[k];
				}

				//for(int k=0; k<3; k++)
					//pixelrgb[k]  = threshold(pixelrgb[k] / (float)samples, 0.0f, 2.0f) ;
				//	pixelrgb[k] = pixelrgb[k] / (float)samples;

				//NOTHREAD
				//EnterCriticalSection(&lock);

				changed = true;
				//image->SetPixelRGB(i, j, *(RNRgb*)&pixelrgb);

				globalstate = 15.1;

				pixelcounts[i][j] += samples;
				for(int k=0; k<3; k++)
				{
					globalstate = 15.2;
					pixelsums[i][j][k] += (pixelrgb[k] < 0 ? 0 : pixelrgb[k]);
					//pixelsums[i][j][k] += pixelrgb[k]; 
					globalstate = 15.3;
					globali = i;
					globalj = j;

					pixelvalues[i][j][k] = pow((pixelsums[i][j][k] / pixelcounts[i][j]), 1.0f/2.2f);
					globalstate = 15.4;
				}
				pixelvalues[i][j][3] = 1.0;
				globalstate = 16;
				/*
				pixelcounts[j][i] += 1;
				for(int k=0; k<3; k++)
				{
				pixelsums[j][i][k] += pixelrgb[k];
				pixelvalues[j][i][k] = pow((pixelsums[j][i][k] / pixelcounts[j][i]), 1.0f/2.2f);
				}
				pixelvalues[j][i][3] = 1.0;*/

				/*
				for(int k=0; k<3; k++)
				pixelvalues[j][i][k] = (pixelvalues[j][i][k]*pixelcounts[j][i] + samples*pixelrgb[k]) / (pixelcounts[j][i]+samples) ;
				pixelvalues[j][i][3] = 1.0;
				pixelcounts[j][i] += samples;*/
				
				
				//NOTHREAD
				//LeaveCriticalSection(&lock);


				//if( display )
				//	glutPostRedisplay(); 

			} //for j
		} //for i

		printf(" Elapsed time = %.6f seconds\n", start_time.Elapsed());
		WriteImage(NextFile("Pass").c_str());

		//linesrendered = 0;

		//NOTHREAD
		//if( display )			WaitForSingleObject(pass_barrier_event, INFINITE);	
	
	} //for p
	

	// Print statistics
	if (print_verbose) {
		printf(" Elapsed time = %.6f seconds\n", start_time.Elapsed());
	}
}



static int ParseArgs(int argc, char **argv)
{
	// Check number of arguments
	if ((argc == 2) && (*argv[1] == '-')) {
		printf("Usage: montecarlo scenename [-g gridsize_per_pixel]\n");
		exit(0);

	}

	// Parse arguments
	argc--; argv++;
	while (argc > 0) {
		if ((*argv)[0] == '-') {
			if (!strcmp(*argv, "-r")) { 
				//argv++; argc--; imagewidth = atoi(*argv); 
				//argv++; argc--; imageheight = atoi(*argv); 
			}
			else if (!strcmp(*argv, "-g")) {
				argv++; argc--; grid_dimension = atoi(*argv);
			}
			else if (!strcmp(*argv, "-p")) {
				argv++; argc--; render_passes = atoi(*argv);
			}
			else if (!strcmp(*argv, "-verbose")) print_verbose = TRUE;
			else RNAbort("Invalid program argument: %s", *argv);
			argv++; argc--;
		}
		else {
			if (!scenename) scenename = *argv;
			else if (!imagename) imagename = *argv;
			else RNAbort("Invalid program argument: %s", *argv);
			argv++; argc--;
		}
	}

	// Check scene filename
	if (!scenename) {
		RNFail("You did not specify a scene file name");	
		return FALSE;
	}

	// Check image filename
	if (!imagename) {
		RNFail("You did not specify a image file name");	
		return FALSE;
	}

	// Check image dimensions
	if ((imagewidth < 1) || (imageheight < 1)) {
		RNFail("You entered invalid image dimensions: %d %d", imagewidth, imageheight);
		return FALSE;
	}

	// Return OK status 
	return TRUE;
}

int power_of_two(int x)
{
	for(int i=0; i<32; i++)
		if( x == 1<<i)
			return 1;
	
	return 0;
}

int next_power_of_two(int x)
{
	for(int i=0; i<32; i++)
	{
		if( x <= 1<<i)
			return 1<<i;
	}
	return 2;
}

char* resize( int width, int height, char* pixels )
{
	int pw = next_power_of_two(width);
	int ph = next_power_of_two(height);

	char* pixout = new char[pw*ph*3];
	gluScaleImage(GL_RGB, width, height, GL_UNSIGNED_BYTE, pixels, pw, ph, GL_UNSIGNED_BYTE, pixout );
	return pixout;
}

double rotateX=0.0, rotateY=0.0, rotateZ = 0.0;
double transX=0, transY=0, transZ=0;
void display_handler()
{

	//Find all the lights in the scene
	vector<R3Node*> lights;
	for(int i=0; i<scene->NNodes(); i++)
	{
		R3Node* node = (R3Node*)scene->Node(i);
		if( node->Material()->IsEmissive() )
			lights.push_back(node);
	} 

	/*
	//Node 12 = sphere
	Triangle T;
	//ProjectToSphere( ((R3Sphere*)scene->Node(12)->Shape())->Center(),  (R3Triangle*)(lights[1]->Shape()), T);
	ProjectToSphere( R3Point(0.75, 0.25, 0.45),  (R3Triangle*)(lights[1]->Shape()), T);

	vector<Triangle> tris;
	//R3Vector thenorm(-0.5, -1, 0);
	R3Vector thenorm(0, 0, 1);
	thenorm.Normalize();

	ClipSphericalTriangle(thenorm, T, tris);
*/

	static GLfloat light0_position[] = { 3.0, 4.0, 5.0, 0.0 };
	static GLfloat light1_position[] = { -3.0, -2.0, -3.0, 0.0 };

	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	bool need_delete = false;

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);

	glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
	glLightfv(GL_LIGHT1, GL_POSITION, light1_position);


	glPushMatrix();

	scene->Camera()->Load();
	
	glTranslatef(scene->BBox().Centroid().X() + transX, scene->BBox().Centroid().Y() + transY, scene->BBox().Centroid().Z() + transZ);
	glRotatef(rotateX, scene->Camera()->Left().X(), scene->Camera()->Left().Y(), scene->Camera()->Left().Z());
	glRotatef(rotateY, scene->Camera()->Up().X(), scene->Camera()->Up().Y(), scene->Camera()->Up().Z());
	glRotatef(rotateZ, scene->Camera()->Towards().X(), scene->Camera()->Towards().Y(), scene->Camera()->Towards().Z());
	glTranslatef(-scene->BBox().Centroid().X(), -scene->BBox().Centroid().Y(), -scene->BBox().Centroid().Z());

	scene->Draw();
	

	/*
	R3Vector outgoing(1,1,1);
	outgoing.Normalize();	
	R3Vector &center = *(R3Vector*)&(((R3Sphere*)scene->Node(12)->Shape())->Center());		
	
	glLineWidth(1.0);
	glBegin(GL_LINES);
	for(int i=0; i<200; i++)
	{
		R3Vector sample;
		SamplePhongLobe(outgoing, 500, sample);
		
		//SampleSphericalTriangle(T, sample);

		//R3Vector up(0, 0, 1);
		//sample_hemisphere(up, sample);

		R3Vector pos;
		pos = center + sample;

		glVertex3fv( (float*)&center );
		glVertex3fv( (float*)&pos );
	}
	glEnd();

	glLineWidth(5.0);
	glBegin(GL_LINES);
	glColor3f(0, 0, 1);
	glVertex3fv( (float*)&center );
	glVertex3fv( (float*)&(center + outgoing) );

	glEnd();
	*/
		/*
	glLineWidth(5.0);
	glBegin(GL_LINES);
	for(int j=0; j<tris.size(); j++)
	{
		Triangle &U = tris[j];
		for(int i=0; i<3; i++)
		{
			//R3Vector &center = *(R3Vector*)&(((R3Sphere*)scene->Node(12)->Shape())->Center());
			R3Vector center(0.75, 0.25, 0.45);
			R3Vector pos;
			pos = center + U.vtx[i];

			glVertex3fv( (float*)&center );
			glVertex3fv( (float*)&pos );
		}
	}
	glEnd();
*/

	/*
	glDisable(GL_LIGHTING);
	glPushMatrix();
	glTranslatef(0.5, 0.5, 0.5);
	glBegin(GL_LINES);
	glColor3f(1,0,0);
	glVertex3f(0,0,0);
	glVertex3f(1,0,0);

	glColor3f(0,1,0);
	glVertex3f(0,0,0);
	glVertex3f(0,1,0);

	glColor3f(0,0,1);
	glVertex3f(0,0,0);
	glVertex3f(0,0,1);
	glEnd();
	glPopMatrix();

	glBegin(GL_LINES);
	glColor3f(0,1,1);
	glVertex3f(0.5, 0.5, 0.5);
	glVertex3fv( (float*)&(R3Vector(0.5, 0.5, 0.5) + thenorm) );
	glEnd();
	*/
	
	

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	
	glPopMatrix();

	//glutSwapBuffers();
	//return;

	if( display_tex )
	{

		
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluOrtho2D(-1, 1, -1, 1);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		
		glEnable(GL_ALPHA_TEST);
		glAlphaFunc(GL_EQUAL, 1.0);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_CULL_FACE);

		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, texId);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		
		
		glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex2f(-1, -1);

			//glTexCoord2f(0,1);
			glTexCoord2f(0, (float)imagewidth/actualwidth);
			glVertex2f(+1, -1);

			glTexCoord2f((float)imageheight/actualheight, (float)imagewidth/actualwidth);
			//glTexCoord2f(1,1);
			glVertex2f(+1, +1);
						
			//glTexCoord2f(1,0);
			glTexCoord2f((float)imageheight/actualheight, 0);
			glVertex2f(-1, +1);
		glEnd(); 

		//if( need_delete )
		//	delete [] pixels;

		glDisable(GL_ALPHA_TEST);
		glDisable(GL_TEXTURE_2D);
	}
	else
	{
		//NOTHREAD
		//EnterCriticalSection(&lock);
		glDrawPixels(image->Width(), image->Height(), GL_RGB, GL_UNSIGNED_BYTE, image->Pixels());
		//NOTHREAD
		//LeaveCriticalSection(&lock);
	}

	glutSwapBuffers();
}

void idle_handler()
{
	bool update = false;

	//Get input
	
	//NOWINDOWS
	/*
	if( GetAsyncKeyState(VK_LEFT) )
	{
		update = true;
		rotateY += 1.0;
	}
	else if( GetAsyncKeyState(VK_RIGHT) )
	{
		update = true;
		rotateY -= 1.0;
	} 
	
	if( GetAsyncKeyState(VK_UP) )
	{
		update = true;
		rotateX += 1.0;
	}
	else if( GetAsyncKeyState(VK_DOWN) )
	{
		update = true;
		rotateX -= 1.0;
	} 
	*/

	/*
	//EnterCriticalSection(&lock);
	if( changed )
	{			
	
		//glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 
		//	actualwidth, 
		//	actualheight,
		//	GL_RGBA, GL_FLOAT, pixelvalues); 

		glPixelStorei(GL_UNPACK_ROW_LENGTH, imagewidth);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 
			imageheight,
			imagewidth,
			GL_RGBA, GL_FLOAT, pixelvalues);
		changed = false;
		glutPostRedisplay();
	}
	//LeaveCriticalSection(&lock);
	*/
	
	int mylinesrendered = linesrendered;
	if( linesdrawn > mylinesrendered )
		linesdrawn = 0;

	if( mylinesrendered > linesdrawn )
	{
		glBindTexture(GL_TEXTURE_2D, texId);
		int count = linesrendered - linesdrawn;
		glPixelStorei(GL_UNPACK_ROW_LENGTH, imagewidth);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 
			0, linesdrawn, 		//x,y offset
			imageheight, count, //width, height
			GL_RGBA, GL_FLOAT, pixelvalues[linesdrawn]);
/*	
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 
			imageheight,
			imagewidth,
			GL_RGBA, GL_FLOAT, pixelvalues);*/
		changed = false;
		linesdrawn = mylinesrendered;
		glutPostRedisplay();
	} 

	////NOTHREAD
	//if( mylinesrendered == imageheight )		PulseEvent(pass_barrier_event);

	if( update )
		glutPostRedisplay();
}



//NOTHREAD
/*
unsigned long WINAPI RenderThread(void* param)
{
	Render(scene, image);
	WriteImage(NextFile("Output").c_str());
	return 0;
}
*/

//NOTHREAD
//HANDLE thread;

int main(int argc, char **argv)
{
	//NOTHREAD
	display = false;

	//If display mode is on, initialize GLUT
	if( display )
	{
		glutInit(&argc, argv);
    	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
		glutInitWindowSize(imagewidth,imageheight);		
		win = glutCreateWindow("CDLib Tester");
		glutDisplayFunc( display_handler );
		glutIdleFunc( idle_handler );

		glGenTextures(1, &texId);
		//NOTHREAD
		//CreateEvent(NULL, false, false, NULL);
	}	

	//NOTHREAD
	//InitializeCriticalSection(&lock);
	NoiseInit();

	// Initialize RING
	if (!R3InitGraphics()) exit(1);

	// Parse program arguments
	if (!ParseArgs(argc, argv)) exit(1);

	// Create input scene
	scene = R3ReadRayFile(scenename);
	if (!scene) exit(1);

	// Create output image and init to zero
	image = new R2Image(imagewidth, imageheight);
	assert(image);
	for(int i=0; i<image->Width(); i++)
		for(int j=0; j<image->Height(); j++)
			image->SetPixelRGB(i, j, RNRgb(0.0, 0.0, 0.0));

	RNSetEpsilon(0.0001);


	//If we are in display mode, then call render in another thread
	if( display )
	{
		glBindTexture(GL_TEXTURE_2D, texId);
	    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR);


		bool need_delete = false;
		char* pixels = (char*)image->Pixels();
		
		if( !power_of_two(image->Width()) || !power_of_two(image->Height()) )
		{
			pixels = resize(image->Width(), image->Height(), pixels);
			need_delete = true;
		}

		/*
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, 
			next_power_of_two(image->Width()), 
			next_power_of_two(image->Height()),
			0, GL_RGB, GL_UNSIGNED_BYTE, pixels);*/

		/*
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, 
			imagewidth, imageheight,
			0, GL_RGBA, GL_FLOAT, pixelvalues);*/

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, 
			actualwidth, actualheight,
			0, GL_RGBA, GL_FLOAT, pixelvalues);

		if( need_delete )
			delete [] pixels;

		static GLfloat lmodel_ambient[] = { 0.3, 0.3, 0.3, 1.0 };
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
		glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
		static GLfloat light0_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
		glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
		glEnable(GL_LIGHT0);
		static GLfloat light1_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
		glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
		glEnable(GL_LIGHT1);

		//NOTHREAD
		//thread = CreateThread(NULL, 0, RenderThread, NULL, CREATE_SUSPENDED, NULL);
		//SetThreadPriority(thread, THREAD_PRIORITY_BELOW_NORMAL);
		//ResumeThread(thread);

		glutMainLoop();
	}
	else
	{
		Render(scene, image);
		//image->Write(imagename);
		WriteImage(NextFile("Output").c_str());
		R3StopGraphics();
	}

	// Return success 
	return 0;
}


	/*
	if( false )
	{
		double colorMin[] = { 255.0/255.0, 217.0/255.0, 160.0/255.0 };
		double colorMax[] = { 70.0/255.0, 12.0/255.0, 0/255.0 };

		double nt = NoiseSum(point[0], point[1], point[2], 4);	
		
		for(int k=0; k<3; k++)
			weightout[k] = weightin[k] * ((1-nt)*colorMin[k]+nt*colorMax[k]) * cosND / (1-P_emit);
	}
	else
		for(int k=0; k<3; k++)
			///weightout[k] = weightin[k] * brdf.Diffuse()[k] * cosND / (1-P_emit);
			weightout[k] = weightin[k] * brdf.Diffuse()[k] * cosND / (P_ray);
	*/

